/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of libcftp - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License (see COPYING.LGPL 
 * for details).
 *
 * $Id: common.c,v 1.5 2000/05/05 22:20:26 malekith Exp $
 */

#include "config.h"
#include "clib_int.h"
#include "version.h"

#include <sys/types.h>

#ifdef WINDA
#include <winsock.h>
#else /* !WINDA */
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#endif /* WINDA */

#include <string.h>
#include <stdlib.h>
#include <errno.h>

u32 xcrc32(const void *s, int len);

#ifdef USE_ZLIB
static struct packet *z_decompress(CTX *ctx, struct packet *in);
static void z_finish_decompress(CTX *ctx);
#endif /* USE_ZLIB */

int set_error(CTX *ctx, char *str)
{
	xfree(ctx->error);
	ctx->error = str;
	return -1;
}

int set_sock_error(CTX *ctx, const char *string)
{
#ifdef WINDA
	xfree(ctx->error);
	ctx->error = xprintf("%s: %s", string, get_wsa_error());
#else
	int k;

	k = errno;
	xfree(ctx->error);
	ctx->error = xprintf("%s: %s", string, strerror(k));
#endif
	ctx->u->dead = 1; /* all socket level errors are treated as fatal */
	return -1;
}

int read_l(SOCKET fd, void *ptr, int size)
{
	int n, r;
	
	if (fd == -1)
		return -1;
	if (!size)
		return size;
	for (n = size; n; n -= r, ptr = (byte*)ptr + r)
		if ((r = recv(fd, ptr, n, 0)) <= 0)
			return -1;
	return size;
}

int write_l(SOCKET fd, const void *ptr, int size)
{
	int n, r;
	
	if (fd == -1)
		return -1;
	if (!size)
		return size;
	for (n = size; n; n -= r, ptr = (byte*)ptr + r)
		if ((r = send(fd, ptr, n, 0)) <= 0)
			return -1;
	return size;
}
int send_packet(CTX *ctx, int id, void *data, int len)
{
	struct packet *buf;
	int n, ntotal;

	/* n = len + sizeof(struct packet) - 1 - 4; */
	n = len + 6;
	
	if (ctx->u->encrypting && n % ctx->u->e->blocklen)
		ntotal = n + ctx->u->e->blocklen - n % ctx->u->e->blocklen;
	else
		ntotal = n;
	buf = (struct packet*)xmalloc(ntotal + 4);
	buf->id = id;
	buf->size = h2n(ntotal);
	buf->crc32 = h2n(xcrc32(data, len));
	if ((buf->pad = ntotal - n) != 0)
		random_get(buf->data + len, buf->pad);
	memcpy(buf->data, data, len);
	if (ctx->u->encrypting)
		ctx->u->e->cipher(ctx->u->e, ((byte*)buf) + 4, ntotal);

	if (write_l(ctx->sock, buf, ntotal + 4) != ntotal + 4) {
		set_sock_error(ctx, "send");
		return -1;
	}

	xfree0(buf, ntotal + 4);
	return 0;
}

int proto_err(CTX *ctx, const char *msg)
{
	return set_error(ctx, xprintf("error in protocol: %s", msg));
}

struct packet *recv_packet(CTX *ctx)
{
	struct packet *buf;
	u32 len;

	if (read_l(ctx->sock, &len, 4) != 4) {
		set_sock_error(ctx, "recv(4)");
		return 0;
	}
	
	len = n2h(len);
	buf = (struct packet *) xmalloc(len + 4 + 1);
	/* buf->size = len - (sizeof(struct packet) - 1 - 4); */
	buf->size = len - 6;
	
	if (read_l(ctx->sock, ((byte*)buf)+4, len) != (int)len) {
		set_sock_error(ctx, "recv");
		goto shit; 
	}
		
	if (ctx->u->encrypting && len % ctx->u->d->blocklen) {
		proto_err(ctx, "input packet not aligned");
		goto shit;
	}

	if (ctx->u->encrypting)
		ctx->u->d->decipher(ctx->u->d,((byte*)buf)+4, len);
	
	buf->size -= buf->pad;
	buf->pad = 0;
	buf->data[buf->size] = 0;
	
	if (xcrc32(buf->data, buf->size) != (u32)n2h(buf->crc32)) {
		proto_err(ctx, "bad crc32"); 
		goto shit;
	}
	
#ifdef USE_ZLIB
	switch (buf->id) {
	case PKT_ZFILE:
		{	
		struct packet *b2;
		b2 = z_decompress(ctx, buf);
		if (!b2)
			goto shit;
		buf = b2;
		}
		break;
	case PKT_EOF:
	case PKT_ABORT:
		z_finish_decompress(ctx);
		break;
	}
#endif /* USE_ZLIB */

	return buf;

shit:
	xfree0(buf, len + 4);
	return 0;
}

struct packet *recv_packet_x(CTX *ctx, int id)
{
	struct packet *p;

	p = recv_packet(ctx);
	if (p && p->id != id) {
		proto_err(ctx, "got unexpected packet ID");
		free_packet(p);
		p = 0;
	}
	return p;
}

void free_packet(struct packet *p)
{
	if (!p)
		return;
	/* xfree0(p, p->size + sizeof(struct packet)); */
	xfree0(p, p->size + 10);
}

int check_in(CTX *ctx)
{
	fd_set read;
	struct timeval tv = {0, 0};
	
	if (ctx->sock == INVALID_SOCKET)
		return 0;
		
	FD_ZERO(&read);
	FD_SET(ctx->sock, &read);
	
	return select(ctx->sock+1, &read, 0, 0, &tv) != 0;
}

#ifdef USE_ZLIB	
static void *z_alloc(void *h, u32 n, u32 m)
{
	(void)h;
	return xmalloc(n*m);
}
static void z_free(void *h, void *ptr)
{
	(void)h;
	xfree(ptr);
}
#endif /* USE_ZLIB */

int z_compress(CTX *ctx, const void* buf, int len)
{
	struct zc_state *zc;
	zc = &ctx->u->zc;
	
	if (len == 0)
		return 0;
		
	
#ifdef USE_ZLIB	
	if (zc->compress) {
		zc->s.avail_in = len;
		zc->s.next_in = (void*)buf;
		while (zc->s.avail_in) {
			zc->s.avail_out = CHUNK_SIZE - zc->chunk_pos;
			zc->s.next_out = zc->buf + zc->chunk_pos;
			if (deflate(&zc->s, 0) != Z_OK) {
				set_error(ctx, xprintf("deflate returned error"));
				return -1;
			}
			if (zc->s.avail_out == 0) {
				send_packet(ctx, zc->pktid, zc->buf,
					    CHUNK_SIZE);
				zc->chunk_pos = 0;
			} else
				zc->chunk_pos = CHUNK_SIZE - zc->s.avail_out;
		}
	} else
#endif /* USE_ZLIB */
	{
		while (len > 0) {
			int n = CHUNK_SIZE - zc->chunk_pos;
			if (n == 0) {
				send_packet(ctx, zc->pktid, zc->buf, CHUNK_SIZE);
				zc->chunk_pos = 0;
				n = CHUNK_SIZE;
			}
			if (len < n)
				n = len;
			memcpy(zc->buf, buf, n);
			len -= n;
			zc->chunk_pos += n;
		}
	}
	return 0;
}

int z_finish_compress(CTX * ctx)
{
	struct zc_state *zc;
	zc = &ctx->u->zc;

#ifdef USE_ZLIB
	if (zc->compress) {
		zc->s.avail_out = CHUNK_SIZE - zc->chunk_pos;
		zc->s.next_out = zc->buf + zc->chunk_pos;
		while (deflate(&zc->s, Z_FINISH) == Z_OK) {
			if (send_packet(ctx, zc->pktid, zc->buf, 
			                 CHUNK_SIZE))
				return -1;
			zc->s.avail_out = CHUNK_SIZE;
			zc->s.next_out = zc->buf;
		}
		if (zc->s.avail_out != CHUNK_SIZE
		    && send_packet(ctx, zc->pktid, zc->buf,
				   CHUNK_SIZE - zc->s.avail_out)) 
			return -1;
		if (deflateEnd(&zc->s) != Z_OK) {
			set_error(ctx,
				  xstrdup("deflateEnd(): returned error"));
			return -1;
		}
	} else
#endif				/* USE_ZLIB */
	{
		if (zc->chunk_pos
		    && send_packet(ctx, zc->pktid, zc->buf,
				   zc->chunk_pos)) 
			return -1;
	}
	return 0;
}

int z_start_compress(CTX *ctx, int pktid, int compress)
{
	struct zc_state *zc;
	zc = &ctx->u->zc;
	
	zc->pktid = pktid;
	zc->chunk_pos = 0;
#ifdef USE_ZLIB
	if ((zc->compress = compress) != 0) {
		zc->s.zalloc = z_alloc;
		zc->s.zfree = z_free;
		zc->s.avail_in = 0;
		zc->s.avail_out = 0;
		zc->s.next_in = 0;
		zc->s.next_out = 0;
		if (deflateInit(&zc->s, compress) != Z_OK)
			return set_error(ctx, xstrdup("deflateInit(): returned error")); 
		if (zc->pktid == PKT_FILE)
			zc->pktid = PKT_ZFILE;
	}
#else /* USE_ZLIB */
	(void)compress;
#endif /* USE_ZLIB */
	return 0;
}


#ifdef USE_ZLIB
static void z_finish_decompress(CTX *ctx)
{
	struct zd_state *zd;
	zd = &ctx->u->zd;
	
	if (zd->initialised) {
		zd->initialised = 0;
		if (zd->pending)
			xfree0(zd->pending, zd->pending_size);
		zd->pending = 0;
		zd->pending_size = 0;
		inflateEnd(&zd->s);
	}
}

static struct packet *z_decompress(CTX *ctx, struct packet *in)
{
	struct packet *out;
	byte *free_me;
	int free_me_size;
	byte buf[CHUNK_SIZE * 5];
	int r;
	struct zd_state *zd;
	
	zd = &ctx->u->zd;

	out = (struct packet *) xmalloc(
/*		sizeof(struct packet) - 1 + sizeof(buf) + 1); */
		10 + sizeof(buf) + 1);
		
	out->size = sizeof(buf);

	zd->csize += in->size;
	
	free_me = (byte *) xmalloc(free_me_size = zd->pending_size + in->size);
	if (zd->pending_size)
		memcpy(free_me, zd->pending, zd->pending_size);
	memcpy(free_me + zd->pending_size, in->data, in->size);
	if (zd->pending_size)
		xfree0(zd->pending, zd->pending_size);
	zd->pending = 0;
	zd->pending_size = 0;

	zd->s.next_in = free_me;
	zd->s.avail_in = free_me_size;
	zd->s.avail_out = sizeof(buf);
	zd->s.next_out = buf;
	zd->s.zalloc = z_alloc;
	zd->s.zfree = z_free;

	if (!zd->initialised)
		inflateInit(&zd->s);
	zd->initialised = 1;

	while ((r = inflate(&zd->s, Z_SYNC_FLUSH)) == Z_OK || r == Z_STREAM_END) {
		struct packet *np;
		memcpy(out->data + out->size - sizeof(buf), buf,
		       sizeof(buf));
		if (zd->s.avail_out != 0)
			break;
		np = (struct packet *) xmalloc(out->size + sizeof(buf) +
					/*      sizeof(struct packet)); */
						10 + 1);
		memcpy(np->data, out->data, out->size);
		np->size = out->size + sizeof(buf);
		xfree0(out, out->size + 10 + 1);
		out = np;
		zd->s.avail_out = sizeof(buf);
		zd->s.next_out = buf;
	}

	if (zd->s.avail_in)
		memcpy((zd->pending = 
			(byte *) xmalloc(zd->pending_size = zd->s.avail_in)), 
			zd->s.next_in, zd->s.avail_in);
	xfree0(free_me, free_me_size);
	out->id = in->id;
	out->size -= zd->s.avail_out;
	out->crc32 = 0;
	out->pad = 0;
	free_packet(in);
	
	if (r == Z_STREAM_END) {
		z_finish_decompress(ctx);
	} else if (r != Z_OK) {
		set_error(ctx, xprintf("inflate() returned error"));
		return 0;
	}
	switch (out->id) {
	case PKT_ZFILE:
		out->id = PKT_FILE; break;
	}
	return out;
}
#endif /* USE_ZLIB */

int z_decompress_size(CTX *ctx)
{
#ifdef USE_ZLIB 
	int r;
	
	r = ctx->u->zd.csize;
	ctx->u->zd.csize = 0;
	return r;
#else /* !USE_ZLIB */

	(void)ctx;
	return 0;
#endif /* USE_ZLIB */
}


/* ---------------> new */
int version_agreement(CTX * ctx)
{
	const char *ver = "CFTP-" PROTOCOL_MAJOR "." PROTOCOL_MINOR
#ifdef __unix__
	    "u"
#endif
#ifdef USE_ZLIB
	    "z"
#endif
	    "-" SOFTWARE_VERSION "\n";
	char buf[1024] = { 0 };
	int i = 0;
	int p1, p2;

	write_l(ctx->sock, ver, strlen(ver));

	do {
		read_l(ctx->sock, &buf[i], 1);
	} while (buf[i++] != '\n' && i < 1000);

	buf[--i] = 0;
	ctx->can_deflate = 0;
	ctx->unix_fs = 0;

	if (strncmp(buf, "CFTP-", 5) != 0)
		return set_error(ctx,
				 xprintf("didn't recive CFTP- header [%s]",
					 buf));

	i = 5;
	p1 = atoi(buf + i);
	while (buf[i] >= '0' && buf[i] <= '9')
		i++;
	if (buf[i] != '.')
		return set_error(ctx,xprintf("bad CFTP- header "
					     "format (no .) [%s]", buf));
	i++;
	p2 = atoi(buf + i);
	while (buf[i] >= '0' && buf[i] <= '9')
		i++;

	while (buf[i] && buf[i] != '-')
		switch (buf[i++]) {
		case 'z':
			ctx->can_deflate = 1;
			break;
		case 'u':
			ctx->unix_fs = 1;
			break;
		default:
			break;
		}

	ctx->software_version = xstrdup(buf + i);
#ifndef USE_ZLIB
	ctx->can_deflate = 0;
#endif				/* USE_ZLIB */
	(void)p1;
	(void)p2;
	return 0;
}

CTX *clib_alloc()
{
	CTX *ctx;
	
	ctx = xmalloc0(sizeof(CTX));
	ctx->sock = INVALID_SOCKET;
	ctx->u = xmalloc0(sizeof(struct clib_int));
	ctx->u->fwd_local = INVALID_SOCKET;
	return ctx;
}

void clib_free(CTX *ctx)
{
	xfree(ctx->software_version);
	xfree(ctx->error);
	xfree(ctx->host);
	xfree(ctx->message);
	
	if (ctx->sock != INVALID_SOCKET)
		closesocket(ctx->sock);
	if (ctx->u->fwd_local != INVALID_SOCKET)
		closesocket(ctx->u->fwd_local);
	if (ctx->u->e)
		ctx->u->e->free(ctx->u->e);
	if (ctx->u->d)
		ctx->u->d->free(ctx->u->d);
	xfree(ctx->u->hostline);
	xfree(ctx->u->key);
}
