/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of libcftp - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License (see COPYING.LGPL 
 * for details).
 *
 * $Id: cftp.h,v 1.4 2000/04/23 23:07:41 malekith Exp $
 */

/* all routines that returns int, return 0 for success 
 * non-0 on error, negative values for fatal errors */

#ifdef WINDA
#include <windows.h>
#include <winsock.h>
#endif

struct clib_int;
struct clib_state {
	SOCKET sock;
	
	/* last error */
	char *error;
	
	/* if we've got access -- this is motd, otherwise it's reason why
	 * we have access denied */
	char *message;
	
	/* info about server */
	char *host;
	char *software_version;
	int unix_fs, can_deflate;
	
	/* internal state, do not mess */
	struct clib_int *u;
};


/* for unix systems etc_dir should be sth like /home/user/.cftp 
 * this dir is later referanced (in this file) as '~'
 * if you don't change format of any file written there, and
 * you are writing unix app, you probably should use it, since
 * all clients should use the same database of known hosts */
void clib_init(const char *etc_dir);

/* probably atexit(clib_close) is just fine */
void clib_close(void);

/* allocates context for later use */
struct clib_state *clib_alloc(void);

/* retval:
 *  0 - ok, connected
 * -1 - oops, look at ctx->error
 * connect() to server
 */
int clib_open1(struct clib_state *ctx, const char *host, int port);

/* retval;
 *   0 - ok, host found in ~/known_hosts
 *   1 - new host, ask user whatever to continue
 *   2 - ALERT! host id changed !
 *  -1 - other error
 * get RSA key from server and check it, cipher isn't set up with server in 
 * this step, but session keys are generated
 */
int clib_open2(struct clib_state *ctx, const char *cipher);

/* retval:
 *   0 - for sucess, check ctx->error for info to be displayed anyway
 *  -1 - for error in key exchange
 *  -2 - for error in cookie exchange
 * send cipher to server and check connection
 */
int clib_open3(struct clib_state *ctx);

/* retval:
 *  0 - ok
 * -1 - oops
 * authentificate
 * this optional, and needn't be called when we only want anon forward
 */
int clib_open4(struct clib_state *ctx, const char *user, const char *passwd);

/* close the connection, clear the keys, free ctx, call this after you
 * you cancelled after first or second open, or you simply want to close
 * conection */
void clib_free(struct clib_state *ctx);

struct clib_stat {
	/*
	 * MODE is defined:
	 * 0007777 is permision mask (drwxrwxrwx) with suid bits
	 * 0007000 this mask is used for setuid, setgid and sticky bit
	 *   filesystems not supporting more then one user should use 
	 *   OWNER permisions (with mask == 00700)
	 * Higher bits AREN'T same as these from <sys/stat.h> !
	 * these are defined below
	 * MODE == -1 marks end of ls block
	 */
	u32 mode;
	u32 mtime;	/* modification time in seconds since Epoch */
	u32 size;	/* size of file in bytes */
	u32 nlink;	/* for human reader only, don't use in client's code */
	char user[32];
	char group[32];
	char name[256];
	char symlink[1024]; /* if (symlink[0]) { file is symlink } */
	/* added to make life easier */
	int isdir;
	char text_mode[11]; /* human readable mode (-rw-r--r--), 10 chars */
};
enum {
	CLIB_MODE_DIRECTORY	= 010000,
	CLIB_MODE_SYMLINK	= 020000,
};

/* retval:
 *   0 - sucess
 *  -1 - error
 *   1 - eof
 * on eof mode is set to -1 & name[0] = 0
 */
int clib_ls(struct clib_state *ctx, struct clib_stat *buf);

/* ask server to send file NAME 
 * if (whole) { startpos & nbytes are ignored }
 * if (defalte)  { try to request compression } 
 * if (mode) { *mode = mode of file being sent }
 * if (size) { *size = how much bytes will be sent }
 * RETVAL:
 *   0 - ok
 *  -1 - send error
 *  -2 - bad response
 *   1 - server error, see ctx->error
 */
int clib_get(struct clib_state *ctx, 
	     const char *name, bool whole, bool deflate, int startpos, 
	     int nbytes, int *mode, int *size, char **owner, char **group);
/* read data from server, they 'll be placed in malloc()ed *buf, size
 * of them in *size (which musn't be 0)
 * RETVAL:
 *   0 - ok
 *   1 - eof
 *   2 - aborted
 *  -1 - error
 */
int clib_read(struct clib_state *ctx, void **buf, int *size);
/* send PKT_ABORT to server
 * this may take some time to stop server, so you should call clib_read
 * until it return 1 or 2.
 * RETVAL:
 *  0 - ok
 * -1 - no file is being recvied
 * -2 - send error
 *  1 - already aborting
 */
int clib_abort_get(struct clib_state *ctx);

/* for flags arg for clib_stor() */
enum {
	CLIB_STOR_TRUNC	= 0x00000001, /* null file first */
	CLIB_STOR_EXCL	= 0x00000002, /* don't clobber existing file */
	CLIB_STOR_APPEND= 0x00000004, /* if used, startpos is ignored */
};
/* ask server to store file NAME
 * if mode is not known to application (e.g. it works in poor filesystem
 * enviroment w/o permisions) it should use mode of 0644
 * NBYTES param must be supplaied, you can't send more ...
 * defalte is done transparently, and you only need to write data with
 * clib_write. if server and/or client doesn't support defalte - it's
 * ignored.
 * RETVAL:
 *   0 - ok
 *  -1 - send error
 *  -2 - bad response
 *  -3 - problems with zlib
 *   1 - server error, see ctx->error
 */
int clib_stor(struct clib_state *ctx, const char *name, 
              int flags, int mode, int startpos, int nbytes,
	      bool deflate);
/* 0 - ok, -1 - error */
int clib_write(struct clib_state *ctx, const void *buf, int size);
/* end of store operation, could be used to abort 
 * RETVAL:
 *   0 - ok
 *   1 - aborted
 *   2 - server claims to got PKT_FILE, but didn't get CMD_STOR first 
 *       (clib_write() w/o clib_stor() first)
 *   3 - server claims to got too many bytes
 *   4 - server got write error (look at ->error)
 *  -1 - send error
 *  -2 - recv error
 *  -3 - unexpected response
 */
int clib_stor_finish(struct clib_state *ctx);

/* executes info command, wich returns some general info on cryptoFTP,
 * returned string must be free()ed */
char *clib_info(struct clib_state *ctx);

/* executes pwd [print working directory] command, string must be free()ed */
char *clib_pwd(struct clib_state *ctx);

/* this accepts '..' for dir up, path separator is of course '/' */
int clib_cd(struct clib_state *ctx, const char *dir);

/* remove file */
int clib_rm(struct clib_state *ctx, const char *filename);

/* remove file/directory and all entries in it (rm -r filename) */
int clib_rmr(struct clib_state *ctx, const char *filename);

/* make directory */
int clib_mkdir(struct clib_state *ctx, const char *filename);

/* remove directory */
int clib_rmdir(struct clib_state *ctx, const char *filename);

/* move file */
int clib_mv(struct clib_state *ctx, const char *oldname, const char *newname);

/* ask server to forward HOST:PORT. open a socket connection to it. */
int clib_forward(struct clib_state *ctx, int port, const char *host, int local);
/* call this in child process/thread, returns string describing whatever
 * did it fail, result must me free()d */
char *clib_forward_forked(struct clib_state *ctx);

/* 0 for success, -1 when connection is no longer usable and should be 
 * closed with clib_free(ctx) */
int clib_clr_error(struct clib_state *ctx);
