/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * this is free software. you can copy and distriute it under terms
 * of GNU General Public License (see COPYING.GPL)
 *
 * $Id: auth.c,v 1.5 2000/05/04 19:09:35 malekith Exp $
 */

#define _XOPEN_SOURCE

#include "config.h"
#include "server.h"

#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifndef BRAINDEAD
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#endif

#ifdef USE_SHADOW
#include <shadow.h>
#endif /* USE_SHADOW */

#ifdef USE_PAM
#include <security/pam_appl.h>
#endif /* USE_PAM */

#define LI(a) (conf_get_bool("VerboseLoginInvalid", 0) ? a : "login invalid")

/* malloced strings */
char *user_home, *user_name, *user_shell;
int user_id, user_gid;

/* bitmask:
 * 1 - invalid (not listed in /etc/shells)
 * 2 - fake (/mail/ || /noshell/ || /null/)
 */
int shell_type;


#ifdef USE_PAM
/* code based on ProFTPD mod_pam module by Habeeb J. Dihu 
 * aka MacGyver <macgyver@tos.net> */
static char *	 pam_pass 		= (char *)0;
static int	 pam_conv_error		= 0;

static int pam_exchange(int num_msg,
			const struct pam_message **msg,
			struct pam_response **resp, void *appdata_ptr)
{
	int i;
	struct pam_response *response = NULL;

	(void)appdata_ptr;
	response = xmalloc(sizeof(struct pam_response) * num_msg);

	for (i = 0; i < num_msg; i++) {
		response[i].resp_retcode = PAM_SUCCESS;

		switch (msg[i]->msg_style) {
		case PAM_PROMPT_ECHO_ON:
			response[i].resp = xstrdup(user_name);
			/* PAM frees resp */
			break;

		case PAM_PROMPT_ECHO_OFF:
			response[i].resp = xstrdup(pam_pass);
			/* PAM frees resp */
			break;

		case PAM_TEXT_INFO:
		case PAM_ERROR_MSG:
			/* ignore it, but pam still wants a NULL response... */
			response[i].resp = NULL;
			break;

		default:
			/* Must be an error of some sort... */
			xfree(response);
			pam_conv_error = 1;
			return PAM_CONV_ERR;
		}
	}

	*resp = response;
	return PAM_SUCCESS;
}

static struct pam_conv pam_conv = {
	pam_exchange,
	NULL
};

char *pam_auth(const char *pass)
{
	int err = 0;
	pam_handle_t *pamh = NULL;
	char *ret = 0;
	
	/* Allocate our entries...we don't free this because PAM does this for us.
	 */
	pam_pass = xstrdup(pass);
	
	err = pam_start("cftp", user_name, &pam_conv, &pamh);

	/* Authenticate, and get any credentials as needed.
	 */
	if (err == PAM_SUCCESS)
		switch ((err = pam_authenticate(pamh, PAM_SILENT))) {
		case PAM_SUCCESS:
			ret = 0; break;
		case PAM_AUTH_ERR:
		case PAM_USER_UNKNOWN:
			ret = "login invalid"; break;
		default:
			ret = "error1"; break;
		}
	else
		ret = "error2";
	
	if (pam_conv_error != 0)
		ret = "error3";

	pam_set_item(pamh, PAM_TTY, "bogus");
	
	if (ret == 0)
		switch ((err = pam_acct_mgmt(pamh, PAM_SILENT))) {
#ifdef PAM_AUTHTOKEN_REQD
		case PAM_AUTHTOKEN_REQD:
			ret = "change your passwd -- it's expired"; break;
#endif				/* PAM_AUTHTOKEN_REQD */
		case PAM_ACCT_EXPIRED:
			ret = "your account has expired"; break;
		case PAM_PERM_DENIED:
			ret = "you can't access system at this time"; break;
		case PAM_SUCCESS: ret = 0; break;
		default: ret = "error4"; break;
		}
		
	if (ret == 0)
		switch ((err = pam_setcred(pamh, PAM_ESTABLISH_CRED))) {
		case PAM_SUCCESS: ret = 0; break;
		case PAM_CRED_EXPIRED: ret = "cred expired"; break;
		default: ret = "error5"; break;
		}

	if (ret)
		notice("PAM(%s): %s [%s]", user_name, pam_strerror(pamh, err), ret);
		
	pam_end(pamh, err);
	
	return ret;
}
#endif

static const char *validate_user()
{
	if (user_id == 0 && !conf_get_bool("AllowRootLogin", 0))
		return LI("you're root");
	else if (user_id < 100 && !conf_get_bool("AllowSystemUID", 0))
		return LI("you're not regular user");
	if ((shell_type & 1) && conf_get_bool("RequireValidShell", 0))
		return "you don't have valid shell -- contact sysadmin";
	if ((shell_type & 2) && conf_get_bool("RequireNonFakeShell", 0))
		return "you have only fake shell -- contact sysadmin";
	return 0;
}

#ifndef BRAINDEAD
static void free0(char *p)
{
	char *f;
	if ((f = p) == 0)
		return;
	while (*p)
		*p++ = 0;
	xfree(f);
}

static void check_shell()
{
	char *sh;
	
	shell_type = 1;
	if (strstr(user_shell, "mail") || strstr(user_shell, "noshell") ||
	    strstr(user_shell, "null"))
	    	shell_type |= 2;
		
	setusershell();
	while ((sh = getusershell()))
		if (strcmp(user_shell, sh) == 0) {
			shell_type &= ~1;
			break;
		}
	endusershell();
}

/* 
 * the PKT_AUTH data should be:
 * user \0 password \0
 * (if password is empty : user \0 \0)
 */
const char *auth_user(char buf[], int buflen)
{
	int i;
	struct passwd *pwd;
	const char *fail_anyway = 0;
	char *user_pass;
	
	for (i = 0; i < buflen - 1 && buf[i]; i++);
	
	if (i == buflen - 1)
		return "error in protocol";
	i++;
	
	/* now we have user in &buf[0] and pass in &buf[i] */
	pwd = getpwnam(buf);
	if (!pwd)
		return LI("you don't exist -- go away");
	user_name = xstrdup(pwd->pw_name);
	user_home = xstrdup(pwd->pw_dir);
	user_shell = xstrdup(pwd->pw_shell);
	user_id = pwd->pw_uid;
	user_gid = pwd->pw_gid;
	check_shell();

	if (pwd->pw_passwd[0] == '!' || pwd->pw_passwd[0] == '*')
		fail_anyway = "your account is disabled";
		
#ifdef USE_SHADOW
	if (strcmp(pwd->pw_passwd, "x") == 0) {
		struct spwd *sp;
		int day;
		
		sp = getspnam(user_name);
		if (sp == 0)
			return LI("you're not in shadow");
		day = time(0) / (60 * 60 * 24);
		if (sp->sp_expire > 0 && day >= sp->sp_expire)
			fail_anyway = "your account has expired";
		else if (sp->sp_inact > 0 && day >= sp->sp_inact)
			fail_anyway = "your account is inactive";
		
		user_pass = xstrdup(sp->sp_pwdp);
		endspent();
	} else
#endif
		user_pass = xstrdup(pwd->pw_passwd);
	/* endpwent(); */

#ifdef USE_PAM
	fail_anyway = pam_auth(buf + i);
#else
	if (strcmp(crypt(buf + i, user_pass), user_pass) != 0)
		fail_anyway = LI("bad password");
#endif
	free0(user_pass);
	
	if (fail_anyway == 0)
		fail_anyway = validate_user();
		
	return fail_anyway;
}

/* uid -> user (and gid->group) conversion, results are cached to speed 
 * things up ...
 */
struct id2name {
	int id;
	char *name;
	struct id2name *next;
};
struct id2name *i2n_uid, *i2n_gid;

static char *find(struct id2name *p, int id)
{
	for (; p; p = p->next)
		if (p->id == id)
			return p->name;
	return 0;
}

char *get_user(int uid)
{
	char *r;
	struct passwd *pwd;
	struct id2name *x;
	
	r = find(i2n_uid, uid);
	if (r)
		return r;
	pwd = getpwuid(uid);
	
	if (!pwd) r = "[???]";
	else r = pwd->pw_name;

	x = (struct id2name*) xmalloc(sizeof(struct id2name));
	r = x->name = xstrdup(r);
	x->id = uid;
	x->next = i2n_uid;
	i2n_uid = x;
	return r;
}

char *get_group(int gid)
{
	char *r;
	struct group *grp;
	struct id2name *x;
	
	r = find(i2n_uid, gid);
	if (r)
		return r;
	grp = getgrgid(gid);
	
	if (!grp) r = "[???]";
	else r = grp->gr_name;

	x = (struct id2name*) xmalloc(sizeof(struct id2name));
	r = x->name = xstrdup(r);
	x->id = gid;
	x->next = i2n_gid;
	i2n_gid = x;
	return r;
}
#else /* BRAINDEAD */

char *get_user(int uid)
{
	(void)uid;
	return "user";
}

char *get_group(int uid)
{
	(void)uid;
	return "group";
}

/* bugus auth method for single-user OSes (if windoze could be called OS)
 *   USER  PASS     UID    HOME
 *   user  secret   1000   /
 *   test  tset     1001   /home
 */
const char *auth_user(char buf[], int buflen)
{
	int i;
	const char *fail_anyway = 0;
	
	for (i = 0; i < buflen - 1 && buf[i]; i++);
	
	if (i == buflen - 1)
		return "error in protocol";
	i++;
	
	/* now we have user in &buf[0] and pass in &buf[i] */
	
	if (strcmp(buf, "test") && strcmp(buf, "user"))
		return LI("you don't exist -- go away");
		
	user_name = xstrdup(buf);
	user_id = 1000;
	if (strcmp(user_name, "test") == 0) user_id++;
	user_home = xstrdup(user_id == 1000 ? "/" : "/home");
	user_shell = xstrdup("/bin/sh");
	user_gid = 1000;
	shell_type = 0; /* shell ok */

	if (strcmp(buf + i, user_id == 1000 ? "secret" : "tset") != 0)
		fail_anyway = LI("bad password");

	if (fail_anyway == 0)
		fail_anyway = validate_user();
		
	return fail_anyway;
}
#endif

